from django.core.management.base import BaseCommand
from ...spreadsheetservice import SpreadSheetService
from ...importservice import ImportService


class Command(BaseCommand):
    help = 'Periodically import autoimportable google spreadsheets.'

    def handle(self, *args, **options):
        import_service = ImportService(SpreadSheetService)
        import_service.autoimport_forever()



import time
import sys
import string
import json
import hashlib
from collections import OrderedDict
import logging

from django.utils.timezone import now
from django.utils.six.moves.urllib.parse import urlencode

from django.contrib.contenttypes.models import ContentType

from inputflow.models import Input

from . import models
from .spreadsheetservice import SpreadSheetService


logger = logging.getLogger(__name__)


class ImportService:
    def __init__(self, spreadsheet_service):
        self.spreadsheet_service = spreadsheet_service

    def worksheet_row_to_dict(self, row):
        non_empty = [col for col in row if col != '']
        if len(non_empty) == 0:
            return None
        cols = ['COL${}'.format(letter) for letter in string.ascii_uppercase]
        encode = lambda value: value.encode('utf-8') if sys.version_info.major == 2 else value
        result_params = [(col, encode(value)) for col, value in zip(cols, row) if value != '']
        result = OrderedDict(result_params)
        content = json.dumps(result)
        sha1 = hashlib.sha1(content.encode('utf-8'))
        result['_content_hash'] = sha1.hexdigest()
        return result

    def save_worksheet_row_to_input(self, spreadsheet, row):
        encoded_row = urlencode(row)
        input_row = Input()
        input_row.settings = spreadsheet.input_settings
        input_row.internal_source = True
        input_row.format = 'form'
        input_row.raw_content = encoded_row
        db = input_row._state.db
        content_type = ContentType.objects.db_manager(db).get_for_model(spreadsheet)
        input_row.related_model = content_type
        input_row.related_model_object_id = spreadsheet.pk
        input_row.save()
        spreadsheet.last_imported_row = row['row']
        spreadsheet.last_imported_date = now()
        spreadsheet.save()
        input_row.notify()

    def import_to_input_flow(self, spreadsheet, start_row=None, row_count=400):
        access_token = spreadsheet.access_token
        if access_token.is_expired():
            access_token.do_refresh_token()

        timestamp = now()
        exceeded = models.QuotaExceeded.objects.filter(access_token=access_token,
            created__lte=timestamp, wait_until__gte=timestamp).first()
        if exceeded is not None:
            logger.error("Quota exceeded for access token of spreadsheet pk:%s. %s seconds left.", spreadsheet.pk, exceeded.wait_until - timestamp)
            return 0

        if spreadsheet.input_settings is None:
            logger.warn("spreadsheet pk:{} has no input_settings assigned".format(spreadsheet.pk))
            return None
        row_index = start_row or spreadsheet.get_next_row_number()
        count = 0
        imported = True
        while imported:
            range_str = spreadsheet.get_range_string(row_index, row_count)
            try:
                values = self.spreadsheet_service.get_worksheet_range(spreadsheet.access_token.access_token,
                    spreadsheet.url, range_str)
            except KeyError as e:
                if e.args == ('values',):
                    return count
                raise
            except SpreadSheetService.NotFound:
                logger.error("Trying to fetch data from not found spreadsheet pk:%s", spreadsheet.pk)
                return count
            except SpreadSheetService.QuotaExceeded:
                quota_exceeded = models.QuotaExceeded()
                quota_exceeded.access_token = access_token
                quota_exceeded.save()
                logger.error("Quota exceeded while fetching data from pk:%s", spreadsheet.pk)
                return count
            except SpreadSheetService.Error as err:
                logger.error("Error fetching data from pk:%s: %s", spreadsheet.pk, err)
                return count
            imported = False
            for row in values:
                row = self.worksheet_row_to_dict(row)
                if row is None:
                    break
                row['row'] = row_index
                self.save_worksheet_row_to_input(spreadsheet, row)
                row_index += 1
                count += 1
                imported = True
        return count

    def start_import(self, spreadsheet, row_count=400):
        return self.import_to_input_flow(spreadsheet, spreadsheet.get_start_row_number(), row_count)

    def continue_import(self, spreadsheet, row_count=400):
        return self.import_to_input_flow(spreadsheet, spreadsheet.get_next_row_number(), row_count)

    def autoimport_iteration(self):
        logger.info("Started spreadsheets automatically import.")
        for spreadsheet in models.Spreadsheet.objects.filter(enabled=True, automatically_imported=True):
            try:
                count = self.continue_import(spreadsheet)
                logger.info("Finished importing spreadsheet '{}': {} rows imported.".format(spreadsheet.pk, count))
            except SpreadSheetService.NotFound:
                logger.info("Spreadsheet not found pk:{}.".format(spreadsheet.pk))
                continue
            except SpreadSheetService.PermissionDenied:
                logger.info("Access token has no acces to spreadsheet pk:{}.".format(spreadsheet.pk))
                continue
        logger.info("Finished spreadsheets automatically import.")

    def autoimport_forever(self):
        while True:
            settings = models.Settings.objects.first()
            delta = 30 if settings is None else settings.autoimport_sleep
            if delta == 0:
                logger.info("Autoimport disabled.")
                delta = 30
            else:
                self.autoimport_iteration()
            logger.info("Sleeping for %ss...", delta)
            time.sleep(delta)



# Generated by Django 2.2.7 on 2020-02-28 03:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('inputgooglesheets', '0003_spreadsheet_automatically_imported'),
    ]

    operations = [
        migrations.AddField(
            model_name='settings',
            name='autoimport_sleep',
            field=models.PositiveIntegerField(blank=True, default=30, help_text='Amount of seconds to sleep after each autoimport.', verbose_name='Autoimport sleep'),
        ),
    ]



from django import forms
from django.utils.translation import ugettext
from .spreadsheetservice import SpreadSheetService


class SpreadSheetForm(forms.ModelForm):
    class Meta:
        fields = ('name', 'access_token', 'url', 'sheet_index', 'omit_first_row',
            'automatically_imported', 'input_settings','enabled')

    def clean(self):
        cleaned_data = super(SpreadSheetForm, self).clean()

        # make sure access token is not expired
        access_token = cleaned_data['access_token']
        if access_token.is_expired():
            access_token.do_refresh_token()

        # validate url field
        try:
            url = cleaned_data['url']
            metadata = SpreadSheetService.get_spreadsheet_metadata(access_token.access_token, url)
        except SpreadSheetService.NotFound:
            self.add_error('url', ugettext("Spreadsheet not found."))
            return cleaned_data
        except SpreadSheetService.PermissionDenied:
            self.add_error('url', ugettext("You have no access to this spreadsheet."))
            return cleaned_data
        except SpreadSheetService.Error:
            self.add_error('url', ugettext("Error fetching spreadsheet information."))
            return cleaned_data
        except ValueError:
            self.add_error('url', ugettext("Wrong spreadsheet url."))
            return cleaned_data

        # validate sheet_index field
        try:
            sheet_index = cleaned_data['sheet_index']
            metadata['sheets'][sheet_index]
        except IndexError:
            self.add_error('sheet_index', ugettext("Wrong worksheet index."))

        self.spreadsheet_metadata = metadata

        return cleaned_data

    def save(self, *args, **kwargs):
        sheet_index = self.cleaned_data['sheet_index']
        self.instance.url = self.spreadsheet_metadata['spreadsheetUrl']
        self.instance.uid = self.spreadsheet_metadata['spreadsheetId']
        self.instance.spreadsheet_name = self.spreadsheet_metadata['sheets'][sheet_index]['properties']['title']

        return super(SpreadSheetForm, self).save(*args, **kwargs)



# Generated by Django 2.2.7 on 2020-07-22 05:22

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('inputgooglesheets', '0004_settings_autoimport_sleep'),
    ]

    operations = [
        migrations.AlterField(
            model_name='settings',
            name='autoimport_sleep',
            field=models.PositiveIntegerField(blank=True, default=30, help_text='Amount of seconds to sleep after each autoimport (use 0 to disable autoimport).', verbose_name='Autoimport sleep'),
        ),
        migrations.CreateModel(
            name='QuotaExceeded',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='Created')),
                ('wait_until', models.DateTimeField(editable=False, verbose_name='Waith until')),
                ('access_token', models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE, related_name='quotas_exceeded', to='inputgooglesheets.AccessToken', verbose_name='Access token')),
            ],
            options={
                'verbose_name': 'Quota exceeded',
                'verbose_name_plural': 'Quotas exceeded',
            },
        ),
    ]



import requests


class SpreadSheetService:
    class Error(Exception):
        pass

    class NotAuthenticated(Error):
        pass

    class PermissionDenied(Error):
        pass

    class NotFound(Error):
        pass

    class QuotaExceeded(Error):
        pass

    @staticmethod
    def normalize_spreadsheet_url(url):
        if url.startswith('https://'):
            return url
        return 'https://' + url

    @classmethod
    def extract_spreadsheet_id(cls, url):
        "Extracts the spreadsheets id from its url."
        url = cls.normalize_spreadsheet_url(url)
        if url.startswith('https://docs.google.com/spreadsheets/d/'):
            sub_url = url.replace('https://docs.google.com/spreadsheets/d/', "")
            return sub_url.split('/')[0]
        if url.startswith('https://docs.google.com/a/'):
            sub_url = url.replace('https://docs.google.com/a/', "")
            return sub_url.split('/')[3]
        raise ValueError("Wrong spreadsheet url: {}".format(url))

    @classmethod
    def check_response(cls, response):
        if response.status_code == 401:
            raise cls.NotAuthenticated("Not authenticated or authentication failed.")
        if response.status_code == 404:
            raise cls.NotFound("Google spreadsheet not found.")
        if response.status_code == 403:
            raise cls.PermissionDenied("You have no access to this spreadsheet.")
        if response.status_code == 429:
            error_data = response.json()
            if error_data['error']['status'] == 'RESOURCE_EXHAUSTED':
                raise cls.QuotaExceeded("Quota exceeded")
        if response.status_code == 200:
            return response.json()
        raise cls.Error("Unknown response status code.")

    @classmethod
    def get_spreadsheet_metadata(cls, token, url):
        "Returns the metadata of the spreadsheet in the url."
        spreadsheet_id = cls.extract_spreadsheet_id(url)
        headers = {
            'Authorization': 'Bearer {}'.format(token)
        }
        url = 'https://sheets.googleapis.com/v4/spreadsheets/{}?includeGridData=false'.format(spreadsheet_id)
        response = requests.get(url, headers=headers)
        return cls.check_response(response)

    @classmethod
    def get_worksheet_range(cls, token, url, data_range):
        "Returns the data in the range"
        spreadsheet_id = cls.extract_spreadsheet_id(url)
        headers = {
            'Authorization': 'Bearer {}'.format(token)
        }
        url = 'https://sheets.googleapis.com/v4/spreadsheets/{}/values/{}?majorDimension=ROWS'.format(spreadsheet_id, data_range)
        response = requests.get(url, headers=headers)
        response =  cls.check_response(response)
        try:
            return response['values']
        except KeyError:
            return []



import time
import sys
import string
from collections import OrderedDict
import json
import hashlib
import logging
import requests
import pygsheets.client
from django.utils.six.moves.urllib.parse import urlencode
from django.utils.timezone import now
from inputflow.models import Input
from oauth_clients.models import AccessToken
from . import models
from . import credentials
from . import exceptions


logger = logging.getLogger(__name__)
autoimportLogger = logging.getLogger('googlesheets.autoimport')


class Utils:
    @staticmethod
    def check_spreadsheets_url_format(url):
        "Checks the url parameter is a valid editable google spreadsheets url."
        if not url.starts_with('https://docs.google.com/spreadsheets/d/'):
            raise ValueError("wrong url format")
        if len(url.split('/')) < 6:
            raise ValueError("wrong url format")

    @classmethod
    def get_spreadsheets_uid(cls, url):
        "Returns the spreadsheets uid from a valid editable google spreadsheets url."
        cls.check_spreadsheet_url_format(url)
        return url.split('/')[5]

    @staticmethod
    def get_spreadsheets_endpoint_from_uid(uid):
        "Returns the spreadsheets data endpoint url for the uid parameter."
        return 'https://sheets.googleapis.com/v4/spreadsheets/{}'.format(uid)

    @classmethod
    def get_spreadsheets_endpoint_from_url(cls, url):
        "Returns the spreadsheets data endpoint url from an editable spreadsheets url."
        uid = cls.get_spreadsheets_uid(url)
        return cls.get_spreadsheets_endpoint_from_uid(uid)

    @staticmethod
    def get_refreshed_token(token):
        "Returns a refreshed access token. Refreshes the token only if its expired."
        access_token = token
        if isinstance(token, AccessToken) and token.is_expired():
            token.do_refresh_token()
            access_token = token.access_token
        return access_token

    @classmethod
    def get_authorization_headers(cls, token):
        "Returns a dict with oauth2 authorization headers."
        access_token = cls.get_refreshed_token(token)
        return {'Authorization': 'Bearer {}'.format(access_token)}

    @classmethod
    def get_spreadsheets_data(cls, url, token):
        "Returns spreadsheets data."
        cls.check_url_format(url)

        headers = cls.get_authorization_headers(token)
        sheets_url = cls.get_spreadsheets_endpoint_from_url(url)

        res = requests.get(sheets_url, headers=headers)
        if res.status_code == 404:
            raise LookupError("spreadsheet not found")
        if res.status_code == 403:
            raise exceptions.ForbiddenError("you have no access to the spreadsheet")

        return res.json()

    @staticmethod
    def fetch_token_info(token):
        """
        update token user_id and username if the token belongs to the oauth client
        selected in the settings object.
        """
        settings = models.Settings.objects.filter(oauth_client=token.client).first()
        if settings is not None:
            url = "https://www.googleapis.com/oauth2/v1/tokeninfo?access_token={}"
            url = url.format(token.access_token)
            response = requests.get(url)
            if response.status_code == 200:
                info = response.json()
                token.user_id = info['user_id']
                token.username = info['email']

    @staticmethod
    def get_spreadsheet(spreadsheet_obj):
        creds = credentials.Credentials(spreadsheet_obj.access_token)
        client = pygsheets.client.Client(creds)
        return client.open_by_url(spreadsheet_obj.url)

    @classmethod
    def get_worksheet(cls, spreadsheet_obj):
        spreadsheet = cls.get_spreadsheet(spreadsheet_obj)
        return spreadsheet.worksheets()[spreadsheet_obj.sheet_index]

    @staticmethod
    def worksheet_row_to_dict(row):
        non_empty = [col for col in row if col != '']
        if len(non_empty) == 0:
            return None
        cols = ['COL${}'.format(letter) for letter in string.ascii_uppercase]
        encode = lambda value: value.encode('utf-8') if sys.version_info.major == 2 else value
        result = [(col, encode(value)) for col, value in zip(cols, row) if value != '']
        result = OrderedDict(result)
        content = json.dumps(result)
        sha1 = hashlib.sha1(content.encode('utf-8'))
        result['_content_hash'] = sha1.hexdigest()
        return result

    @staticmethod
    def save_worksheet_row_to_input(spreadsheet_obj, row):
        encoded_row = urlencode(row)
        input_row = Input()
        input_row.settings = spreadsheet_obj.input_settings
        input_row.internal_source = True
        input_row.format = 'form'
        input_row.raw_content = encoded_row
        input_row.save()
        spreadsheet_obj.last_imported_row = row['row']
        spreadsheet_obj.last_imported_date = now()
        spreadsheet_obj.save()
        input_row.notify()

    @classmethod
    def import_to_input_flow(cls, spreadsheet_obj, max_count=50, start_row=None):
        if spreadsheet_obj.input_settings is None:
            return
        row_index = start_row or spreadsheet_obj.get_next_row_number()
        worksheet = cls.get_worksheet(spreadsheet_obj)
        count = 0
        imported = True
        while imported:
            values = worksheet.get_values((row_index, 1), (row_index+max_count-1, worksheet.cols))
            imported = False
            for row in values:
                row = cls.worksheet_row_to_dict(row)
                if row is None:
                    break
                row['row'] = row_index
                cls.save_worksheet_row_to_input(spreadsheet_obj, row)
                row_index += 1
                count += 1
                imported = True
        return count

    @classmethod
    def start_import(cls, spreadsheet_obj, max_count=50):
        return cls.import_to_input_flow(spreadsheet_obj, max_count=max_count,
            start_row=spreadsheet_obj.get_start_row_number())

    @classmethod
    def continue_import(cls, spreadsheet_obj, max_count=50):
        return cls.import_to_input_flow(spreadsheet_obj, max_count=max_count,
            start_row=spreadsheet_obj.get_next_row_number())

    @classmethod
    def autoimport_iteration(cls):
        autoimportLogger.info("Started spreadsheets automatically import.")
        for spreadsheet in models.Spreadsheet.objects.filter(enabled=True, automatically_imported=True):
            count = cls.continue_import(spreadsheet)
            autoimportLogger.info("Finished importing spreadsheet '{}': {} rows imported".format(spreadsheet.pk, count))
        autoimportLogger.info("Finished spreadsheets automatically import.")

    @classmethod
    def autoimport_forever(cls):
        while True:
            settings = models.Settings.objects.first()
            delta = 30 if settings is None else settings.autoimport_sleep
            if delta == 0:
                autoimportLogger.info("Autoimport disabled.")
                delta = 30
            else:
                cls.autoimport_iteration()
            autoimportLogger.info("Sleeping for %ss...", delta)
            time.sleep(delta)
